#
# This file is part of the Ingram Micro CloudBlue Connect EaaS Extension Runner.
#
# Copyright (c) 2022 Ingram Micro. All Rights Reserved.
#
import logging

import anvil.server

from connect.client import ConnectClient

from pkg_resources import iter_entry_points

from connect.eaas.core.logging import ExtensionLogHandler, RequestLogger
from connect.eaas.runner.config import ConfigHelper


logger = logging.getLogger(__name__)


class AnvilApp:
    """
    Handle the lifecycle of an Anvil extension.
    """
    def __init__(self, config: ConfigHelper):
        self._config = config
        self._anvilapp_class = self.get_anvilapp_class()
        self._anvilapp_instance = None
        self._logging_handler = None

    @property
    def should_start(self):
        return self._anvilapp_class is not None

    @property
    def variables(self):
        return self._anvilapp_class.get_variables()

    @property
    def readme(self):
        return self._anvilapp_class.get_descriptor()['readme_url']

    @property
    def changelog(self):
        return self._anvilapp_class.get_descriptor()['changelog_url']

    def start(self):
        logger.info('Create anvil connection...')
        var_name = self._anvilapp_class.get_anvil_key_variable()
        logger.info(f'Anvil key variable name: {var_name}')
        anvil_api_key = self._config.variables.get(var_name)
        if not anvil_api_key:
            logger.error(f'Cannot start Anvil application: variable {var_name} not found!')
            return

        logger.info('Starting anvil server...')
        anvil.server.connect(anvil_api_key)
        logger.info('Anvil server started successfully.')
        self.setup_anvilapp()

    def stop(self):
        logger.info('Stopping anvil server...')
        anvil.server.disconnect()
        logger.info('Anvil server stopped successfully.')

    def get_anvilapp_class(self):
        anvil_class = next(iter_entry_points('connect.eaas.ext', 'anvilapp'), None)
        return anvil_class.load() if anvil_class else None

    def setup_anvilapp(self):
        if not self._anvilapp_instance:
            self._anvilapp_instance = self._anvilapp_class(
                self.get_client(),
                self.get_logger(),
                self._config.variables,
            )
            self._anvilapp_instance.setup_anvil_callables()

    def get_client(self):
        return ConnectClient(
            self._config.api_key,
            endpoint=self._config.get_api_url(),
            use_specs=False,
            max_retries=3,
            default_headers=self._config.get_user_agent(),
            logger=RequestLogger(
                self.get_logger(),
            ),
        )

    def get_logger(self):
        """
        Returns a logger instance configured with the LogZ.io handler.
        This logger will be used by the extension to send logging records
        to the Logz.io service.
        """
        logger = logging.getLogger('eaas.extension')
        if self._logging_handler is None and self._config.logging_api_key is not None:
            self._logging_handler = ExtensionLogHandler(
                self._config.logging_api_key,
                default_extra_fields=self._config.metadata,
            )
            logger.addHandler(self._logging_handler)
        return logger



#
# This file is part of the Ingram Micro CloudBlue Connect EaaS Extension Runner.
#
# Copyright (c) 2021 Ingram Micro. All Rights Reserved.
#
import asyncio
import logging


from connect.eaas.core.proto import (
    Message,
    MessageType,
    SetupRequest,
)
from connect.eaas.runner.workers.base import WorkerBase
from connect.eaas.runner.constants import (
    DELAY_ON_CONNECT_EXCEPTION_SECONDS,
    RESULT_SENDER_MAX_RETRIES,
    RESULT_SENDER_WAIT_GRACE_SECONDS,
)
from connect.eaas.runner.helpers import get_version
from connect.eaas.runner.managers import (
    BackgroundTasksManager,
    InteractiveTasksManager,
    ScheduledTasksManager,
)


logger = logging.getLogger(__name__)


class Worker(WorkerBase):
    """
    The Worker is responsible to handle the websocket connection
    with the server. It will send the extension capabilities to
    the server and wait for tasks that need to be processed using
    the tasks manager.
    """
    def __init__(self, config, handler, runner_type=None):
        super().__init__(config)
        self.runner_type = runner_type
        self.handler = handler
        self.results_queue = asyncio.Queue()
        self.background_manager = BackgroundTasksManager(
            self.config,
            self.handler,
            self.results_queue.put,
        )
        self.interactive_manager = InteractiveTasksManager(
            self.config,
            self.handler,
            self.results_queue.put,
        )
        self.scheduled_manager = ScheduledTasksManager(
            self.config,
            self.handler,
            self.results_queue.put,
        )
        self.ws = None
        self.main_task = None
        self.results_task = None

    @property
    def running_tasks(self):
        return (
            self.background_manager.running_tasks
            + self.interactive_manager.running_tasks
            + self.scheduled_manager.running_tasks
        )

    def get_url(self):
        url = self.config.get_events_ws_url()
        url = f'{url}?running_tasks={self.background_manager.running_tasks}'
        url = f'{url}&running_scheduled_tasks={self.scheduled_manager.running_tasks}'
        if self.runner_type:
            url = f'{url}&runner_type={self.runner_type}'
        return url

    def get_setup_request(self):
        return Message(
            version=2,
            message_type=MessageType.SETUP_REQUEST,
            data=SetupRequest(
                event_subscriptions={
                    event_type: event['statuses'] or []
                    for event_type, event in self.handler.events.items()
                },
                variables=self.handler.variables,
                schedulables=self.handler.schedulables,
                repository={
                    'readme_url': self.handler.readme,
                    'changelog_url': self.handler.changelog,
                },
                runner_version=get_version(),
            ),
        ).dict()

    async def process_message(self, data):
        """
        Process a message received from the websocket server.
        """
        message = Message.deserialize(data)
        if message.message_type == MessageType.SETUP_RESPONSE:
            self.process_setup_response(message.data)
        elif message.message_type == MessageType.TASK:
            await self.process_task(message.data)
        elif message.message_type == MessageType.SHUTDOWN:
            await self.shutdown()

    async def process_task(self, task_data):
        """Send a task to a manager based on task category."""
        logger.info(
            f'received new {task_data.options.task_category} task: {task_data.options.task_id}',
        )
        manager = getattr(self, f'{task_data.options.task_category}_manager')
        await manager.submit(task_data)
        logger.info(f'task {task_data.options.task_id} submitted for processing')

    async def result_sender(self):  # noqa: CCR001
        """
        Dequeues results from the results queue and send it to
        the EaaS backend.
        """
        await self.run_event.wait()
        while True:
            if self.results_queue.empty():
                if not self.run_event.is_set() and self.running_tasks == 0:
                    logger.info('Worker exiting and no more running tasks: exit!')
                    return
                await asyncio.sleep(.5)
                continue
            logger.info(
                f'Current processing status: running={self.running_tasks} '
                f'results={self.results_queue.qsize()}',
            )
            result = await self.results_queue.get()
            logger.info(f'Got a result from queue: {result.options.task_id}')
            retries = 0
            while retries < RESULT_SENDER_MAX_RETRIES:
                try:
                    message = Message(
                        version=2,
                        message_type=MessageType.TASK,
                        data=result,
                    )
                    await self.ensure_connection()
                    await self.send(message.serialize())
                    logger.info(f'Result for task {result.options.task_id} has been sent.')
                    break
                except Exception:
                    logger.warning(
                        f'Attempt {retries} to send results for task '
                        f'{result.options.task_id} has failed.',
                    )
                    retries += 1
                    await asyncio.sleep(DELAY_ON_CONNECT_EXCEPTION_SECONDS)
            else:
                logger.warning(
                    f'Max retries exceeded ({RESULT_SENDER_MAX_RETRIES})'
                    f' for sending results of task {result.options.task_id}',
                )

            if not self.run_event.is_set():
                logger.info(
                    f'Current processing status: running={self.running_tasks} '
                    f'results={self.results_queue.qsize()}',
                )

    async def stopping(self):
        result_timeout = self.config.get_timeout('background') + RESULT_SENDER_WAIT_GRACE_SECONDS
        try:
            await asyncio.wait_for(
                self.results_task,
                timeout=result_timeout,
            )
        except asyncio.TimeoutError:
            logger.error(
                f'Cannot send all results timeout of {result_timeout} exceeded, cancel task',
            )
            self.results_task.cancel()
            try:
                await self.results_task
            except asyncio.CancelledError:
                logger.info('Result sender task has been cancelled')

    async def start(self):
        """
        Start the runner.
        """
        self.results_task = asyncio.create_task(self.result_sender())
        await super().start()



import base64
import json
import logging
from io import BytesIO
from urllib.parse import unquote, urlparse

from connect.eaas.core.proto import (
    HttpRequest,
    HttpResponse,
    Message,
    MessageType,
    WebTask,
)


logger = logging.getLogger(__name__)


class RequestResponseCycle:

    def __init__(self, config, app, task, callback):
        self.config = config
        self.app = app
        self.task = task
        self.callback = callback
        self.response_status = None
        self.response_headers = None
        self.body_parts = BytesIO()

    def build_scope(self):
        parsed = urlparse(self.task.request.url)
        return {
            'type': 'http',
            'http_version': '1.1',
            'method': self.task.request.method.upper(),
            'path': unquote(parsed.path),
            'query_string': parsed.query.encode('utf-8'),
            'headers': self.build_headers(),
        }

    def build_headers(self):
        headers = [
            (k.lower().encode('utf-8'), v.encode('utf-8'))
            for k, v in self.task.request.headers.items()
        ]
        headers.extend(
            [
                (
                    'X-Connect-Api-Gateway-Url'.lower().encode('utf-8'),
                    self.config.get_api_url().encode('utf-8'),
                ),
                (
                    'X-Connect-User-Agent'.lower().encode('utf-8'),
                    self.config.get_user_agent()['User-Agent'].encode('utf-8'),
                ),
                (
                    'X-Connect-Installation-Api-Key'.lower().encode('utf-8'),
                    self.task.options.api_key.encode('utf-8'),
                ),
                (
                    'X-Connect-Installation-Id'.lower().encode('utf-8'),
                    self.task.options.installation_id.encode('utf-8'),
                ),
                (
                    'X-Connect-Logging-Level'.lower().encode('utf-8'),
                    self.config.logging_level.encode('utf-8')
                    if self.config.logging_level else b'DEBUG',
                ),
            ],
        )
        if self.config.logging_api_key is not None:
            headers.extend(
                [
                    (
                        'X-Connect-Logging-Api-Key'.lower().encode('utf-8'),
                        self.config.logging_api_key.encode('utf-8'),
                    ),
                    (
                        'X-Connect-Logging-Metadata'.lower().encode('utf-8'),
                        json.dumps(self.config.metadata).encode('utf-8'),
                    ),
                ],
            )
        return headers

    async def receive(self):
        body = (
            base64.decodebytes(self.task.request.content.encode('utf-8'))
            if self.task.request.content else b''
        )
        message = {
            'type': 'http.request',
            'body': body,
            'more_body': False,
        }

        return message

    def build_response(self, status, headers, body):
        logger.debug(f'body type: {type(body)}')
        log = logger.info if status < 500 else logger.error
        log(
            f'{self.task.request.method.upper()} {self.task.request.url} {status} -  {len(body)}',
        )
        task_response = WebTask(
            options=self.task.options,
            request=HttpRequest(
                method=self.task.request.method.upper(),
                url=self.task.request.url,
                headers={},
            ),
            response=HttpResponse(
                status=status,
                headers=headers,
                content=base64.encodebytes(body).decode('utf-8'),
            ),
        )

        message = Message(
            version=2,
            message_type=MessageType.WEB_TASK,
            data=task_response,
        )
        return message.serialize()

    async def send(self, message):
        if message['type'] == 'http.response.start':
            self.response_status = message['status']
            self.response_headers = {
                header[0].lower().decode(): header[1].decode()
                for header in message.get('headers', [])
            }
        elif message['type'] == 'http.response.body':
            self.body_parts.write(message['body'])
            if message.get('more_body', False):
                return
            await self.callback(
                self.build_response(
                    self.response_status,
                    self.response_headers,
                    self.body_parts.getvalue(),
                ),
            )

    async def __call__(self):
        scope = self.build_scope()
        await self.app(
            scope,
            self.receive,
            self.send,
        )



from connect.eaas.runner.managers.background import BackgroundTasksManager  # noqa
from connect.eaas.runner.managers.interactive import InteractiveTasksManager  # noqa
from connect.eaas.runner.managers.scheduled import ScheduledTasksManager  # noqa



#
# This file is part of the Ingram Micro CloudBlue Connect EaaS Extension Runner.
#
# Copyright (c) 2022 Ingram Micro. All Rights Reserved.
#
import asyncio

from connect.eaas.core.proto import (
    Message,
    MessageType,
    SetupRequest,
)
from connect.eaas.runner.asgi import RequestResponseCycle
from connect.eaas.runner.workers.base import WorkerBase
from connect.eaas.runner.helpers import get_version


class WebWorker(WorkerBase):
    """
    The Worker is responsible to handle the websocket connection
    with the server. It will send the extension capabilities to
    the server and wait for tasks that need to be processed using
    the tasks manager.
    """
    def __init__(self, config, handler):
        super().__init__(config)
        self.handler = handler

    def get_url(self):
        return self.config.get_webapp_ws_url()

    def get_setup_request(self):
        return Message(
            version=2,
            message_type=MessageType.SETUP_REQUEST,
            data=SetupRequest(
                ui_modules=self.handler.ui_modules,
                variables=self.handler.variables,
                repository={
                    'readme_url': self.handler.readme,
                    'changelog_url': self.handler.changelog,
                },
                runner_version=get_version(),
            ),
        ).dict()

    async def stopping(self):
        pass

    async def process_message(self, data):
        message = Message.deserialize(data)
        if message.message_type == MessageType.SETUP_RESPONSE:
            self.process_setup_response(message.data)
        elif message.message_type == MessageType.WEB_TASK:
            await self.process_task(message.data)
        elif message.message_type == MessageType.SHUTDOWN:
            await self.shutdown()

    async def shutdown(self):
        await super().shutdown()

    async def process_task(self, task):
        cycle = RequestResponseCycle(
            self.config,
            self.handler.app,
            task,
            self.send,
        )
        asyncio.create_task(cycle())



#
# This file is part of the Ingram Micro CloudBlue Connect EaaS Extension Runner.
#
# Copyright (c) 2021 Ingram Micro. All Rights Reserved.
#
class EaaSError(Exception):
    pass


class MaintenanceError(EaaSError):
    pass


class CommunicationError(EaaSError):
    pass


class StopBackoffError(EaaSError):
    pass
